// This file is automatically generated by Make.bash.
// Any changes made to this file will be lost.
package cynosurex.library.TetherProxLibrary;
import cynosurex.application.TetherProx.R;

import android.graphics.Point;

import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;

class ParserStateMachine {
  State step(String url) throws MalformedURLException {
    State state = new State(url.toCharArray());
    StateTransition s = new StartState().step(state);
    if (s instanceof ErrorState) {
      throw new MalformedURLException(((ErrorState) s).msg);
    }
    return state;
  }

  static class State {
    char[] chars;
    int p;  // pointer
    Map<URLConstants.Part, Point> parts = new HashMap<URLConstants.Part, Point>();

    State(char[] chars) {
      this.chars = chars;
      p = 0;
    }

    public void setPart(URLConstants.Part s, Point i) {
      parts.put(s, i);
    }

    public void setPartIfNotExist(URLConstants.Part s, Point i) {
      if (!parts.containsKey(s)) {
        parts.put(s, i);
      }
    }

    public boolean hasPart(URLConstants.Part s) {
      return parts.containsKey(s);
    }

    public Point getPart(URLConstants.Part s) {
      return parts.get(s);
    }
  }


  static interface StateTransition {
    StateTransition step(State state);
  }

  static class ErrorState implements StateTransition {
    String msg;

    ErrorState(String msg) {
      this.msg = msg;
    }

    public StateTransition step(State state) {
      return this;
    }
  }

  static class StartState implements StateTransition {

    public StateTransition step(State state) {
      char[] chars = state.chars;
      for (int i = state.p, n = chars.length; i < n; i++) {
        if (isColonSlashSlash(chars, i)) {
          state.setPart(URLConstants.Part.PROTOCOL, new Point(0, i));
          state.p = i + 3;
          break;
        }
        state.p = i;
      }
      if (state.hasPart(URLConstants.Part.PROTOCOL)) {
        return new HostState().step(state);
      }
      return new ErrorState("No :// in url");
    }

    private boolean isColonSlashSlash(char[] chars, int i) {
      return chars[i] == ':' && chars[i + 1] == '/' && chars[i + 2] == '/';
    }
  }

  static class HostState implements StateTransition {

    public StateTransition step(State state) {
      int hostStart = state.p;
      char[] chars = state.chars;

      //=========================================================
      // Hosts can be of the following form:
      // - username@host
      // - username@host:port
      // - username:password@host
      // - username:password@host:port
      // - host:port
      // - host
      //=========================================================
      for (int n = chars.length; state.p < n; state.p++) {
        if (chars[state.p] == ':') {
          //=========================================================
          // Scan ahead and look for @ before encountering eof, / or ?
          // If @ found, we have username:password@host form. Else, this is host:port
          //=========================================================
          int usernameAt = -1;
          for (int i = state.p + 1; i < chars.length; ++i) {
            char c = chars[i];
            if (c == '/' || c == '?') {
              break;
            }
            if (c == '@') {
              usernameAt = i;
              break;
            }
          }

          if (usernameAt > -1) {
            state.setPart(URLConstants.Part.USER, new Point(hostStart, state.p));
            state.setPart(URLConstants.Part.PASSWORD, new Point(state.p + 1, usernameAt));
            hostStart = usernameAt + 1;
            state.p = hostStart;
          } else {
            state.setPartIfNotExist(URLConstants.Part.HOST, new Point(hostStart, state.p));
            return new PortState().step(state);
          }
        } else if (chars[state.p] == '@') {     // likely to be in the form username@host
          state.setPart(URLConstants.Part.USER, new Point(hostStart, state.p));
          hostStart = state.p + 1;
        } else if (chars[state.p] == '/') {     // arrived at path, i.e. host/path
          state.setPartIfNotExist(URLConstants.Part.HOST, new Point(hostStart, state.p));
          return new PathState().step(state);
        } else if (chars[state.p] == '?') {     // arrived at query, i.e. host?query
          state.setPartIfNotExist(URLConstants.Part.HOST, new Point(hostStart, state.p));
          return new QueryState().step(state);
        } else if (chars[state.p] == '#') {     // arrived at ref, i.e. host#ref
          state.setPartIfNotExist(URLConstants.Part.HOST, new Point(hostStart, state.p));
          return new RefState().step(state);
        }
      }

      if ((state.p - hostStart) <= 0) {
        return new ErrorState("Zero-length host");
      }

      state.setPartIfNotExist(URLConstants.Part.HOST, new Point(hostStart, state.p));
      return new PathState().step(state);
    }
  }

  static class PortState implements StateTransition {
    public StateTransition step(State state) {
      char[] chars = state.chars;
      if (state.p >= state.chars.length) return new EndState(state);
      if (chars[state.p] == ':') {
        int start = state.p + 1;
        for (int n = chars.length; state.p < n; state.p++) {
          if (chars[state.p] == '/') {
            state.setPartIfNotExist(URLConstants.Part.PORT, new Point(start, state.p));
            return new PathState().step(state);
          } else if (chars[state.p] == '?') {
            state.setPartIfNotExist(URLConstants.Part.PORT, new Point(start, state.p));
            return new QueryState().step(state);
          }
        }
        state.setPartIfNotExist(URLConstants.Part.PORT, new Point(start, state.p));
      }
      return new PathState().step(state);
    }
  }

  static class PathState implements StateTransition {
    // todo: sessionids

    public StateTransition step(State state) {
      char[] chars = state.chars;
      if (state.p >= state.chars.length) return new EndState(state);

      if (chars[state.p] == '/') {
        int start = state.p;
        for (; state.p < chars.length; ++state.p) {
          char c = chars[state.p];
          if (c == '?') {
            state.setPartIfNotExist(URLConstants.Part.PATH, new Point(start, state.p));
            return new QueryState().step(state);
          } else if (c == '#') {
            state.setPartIfNotExist(URLConstants.Part.PATH, new Point(start, state.p));
            return new RefState().step(state);
          }
        }
        state.setPartIfNotExist(URLConstants.Part.PATH, new Point(start, state.p));
      }
      return new EndState(state);
    }
  }

  static class RefState implements StateTransition {
    public StateTransition step(State state) {
      char[] chars = state.chars;
      if (state.p >= state.chars.length) return new EndState(state);
      if (chars[state.p] == '#') {
        state.p++;
        int start = state.p;
        for (; state.p < chars.length; ++state.p) {
          char c = chars[state.p];
          if (c == '?') {
            state.setPartIfNotExist(URLConstants.Part.REF, new Point(start, state.p));
            return new QueryState().step(state);
          }
        }
        state.setPartIfNotExist(URLConstants.Part.REF, new Point(start, state.p));
      }
      return new EndState(state);
    }
  }

  static class QueryState implements StateTransition {

    public StateTransition step(State state) {
      char[] chars = state.chars;
      if (state.p >= state.chars.length) return new EndState(state);

      if (chars[state.p] == '?') {
        state.setPartIfNotExist(URLConstants.Part.QUERY, new Point(state.p + 1, chars.length));
      }
      return new EndState(state);
    }
  }

  static class EndState implements StateTransition {
    State state;

    EndState(State state) {
      this.state = state;
    }

    public StateTransition step(State state) {
      return this;
    }
  }
}
